import requests_mock
import pytest
from pubtools._content_gateway.push_base import PushBase, CGWError, FetcherDict
from tests.fake_cgw_client import TestClient

try:
    import mock
except ImportError:
    from unittest import mock


@pytest.fixture()
@mock.patch("pubtools._content_gateway.push_base.CGWClient", return_value=TestClient())
def push_base_object(mocked_cgw_client):
    push_base = PushBase("http://fake_host_nmae/test", "foo", "bar")
    return push_base


class TestPushBase:
    def test_create_product_1(self, create_product_data, push_base_object):
        push_base_object.process_product(create_product_data)
        assert push_base_object.cgw_client.create_product.calls == [((mock.ANY, create_product_data["metadata"]), {})]

    def test_create_product_2(self, create_product2_data, push_base_object):
        push_base_object.process_product(create_product2_data)
        assert ((mock.ANY, create_product2_data["metadata"]), {}) in push_base_object.cgw_client.create_product.calls

    def test_create_duplicate_product(self, create_product_data, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_product(create_product_data)
        assert str(exception.value) == "Cannot create new product. Record already present."

    def test_update_product(self, update_product_data, push_base_object):
        push_base_object.process_product(update_product_data)
        assert push_base_object.cgw_client.update_product.calls == [((mock.ANY, update_product_data["metadata"]), {})]

    def test_delete_product(self, delete_product_data, push_base_object):
        deleted_id = push_base_object.process_product(delete_product_data)
        assert push_base_object.cgw_client.delete_product.calls == [((mock.ANY, deleted_id), {})]

    def test_delete_invalid_product(self, delete_product_data, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_product(delete_product_data)
        assert str(exception.value) == "Cannot update/delete the product %s %s, id is not set" % (
            delete_product_data["metadata"]["name"],
            delete_product_data["metadata"]["productCode"],
        )

    def test_create_version_1(self, create_version_data, push_base_object):
        push_base_object.process_version(create_version_data)
        assert push_base_object.cgw_client.create_version.calls == [
            ((mock.ANY, mock.ANY, create_version_data["metadata"]), {})
        ]

    def test_create_version_2(self, create_version2_data, push_base_object):
        push_base_object.process_version(create_version2_data)
        assert (
            (mock.ANY, mock.ANY, create_version2_data["metadata"]),
            {},
        ) in push_base_object.cgw_client.create_version.calls

    def test_create_duplicate_version(self, create_version_data, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_version(create_version_data)
        assert str(exception.value) == "Cannot create new version. Record already present."

    def test_update_version(self, update_version_data, push_base_object):
        push_base_object.process_version(update_version_data)
        assert push_base_object.cgw_client.update_version.calls == [
            ((mock.ANY, mock.ANY, update_version_data["metadata"]), {})
        ]

    def test_delete_version(self, delete_version, push_base_object):
        deleted_id = push_base_object.process_version(delete_version)
        assert push_base_object.cgw_client.delete_version.calls == [((mock.ANY, mock.ANY, deleted_id), {})]

    def test_delete_invalid_version(self, delete_version, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_version(delete_version)
        assert "Cannot update/delete the version name:" in str(exception.value)

    def test_create_version_without_product(self, create_version_without_product, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_version(create_version_without_product)
        assert str(exception.value) == "Product name: %s product code: %s not found" % (
            create_version_without_product.get("metadata")["productName"],
            create_version_without_product.get("metadata")["productCode"],
        )

    def test_create_file_1(self, create_file_data, push_base_object):
        push_base_object.process_file(create_file_data)
        assert push_base_object.cgw_client.create_file.calls == [
            ((mock.ANY, mock.ANY, mock.ANY, create_file_data["metadata"]), {})
        ]

    def test_create_file_2(self, create_file2_data, push_base_object):
        push_base_object.process_file(create_file2_data)
        assert (
            (mock.ANY, mock.ANY, mock.ANY, create_file2_data["metadata"]),
            {},
        ) in push_base_object.cgw_client.create_file.calls

    def test_create_duplicate_file(self, create_file_data, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_file(create_file_data)
        assert str(exception.value) == "Cannot create new file. Record already present."

    def test_update_file(self, update_file_data, push_base_object):
        push_base_object.process_file(update_file_data)
        assert push_base_object.cgw_client.update_file.calls == [
            ((mock.ANY, mock.ANY, mock.ANY, update_file_data["metadata"]), {})
        ]

    def test_delete_file(self, delete_file_data, push_base_object):
        deleted_id = push_base_object.process_file(delete_file_data)
        assert push_base_object.cgw_client.delete_file.calls == [((mock.ANY, mock.ANY, mock.ANY, deleted_id), {})]

    def test_delete_invalid_file(self, delete_file_data, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_file(delete_file_data)
        assert "Cannot update/delete file: product version:" in str(exception.value)

    def test_create_file_without_product(self, create_file_without_product, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_file(create_file_without_product)
        assert str(exception.value) == "Product name: %s product code: %s not found" % (
            create_file_without_product.get("metadata")["productName"],
            create_file_without_product.get("metadata")["productCode"],
        )

    def test_create_file_without_version(self, create_file_without_version, push_base_object):
        with pytest.raises(CGWError) as exception:
            push_base_object.process_file(create_file_without_version)
        assert str(exception.value) == "Product name: %s product code: %s version name: %s not found" % (
            create_file_without_version.get("metadata").get("productName"),
            create_file_without_version.get("metadata")["productCode"],
            create_file_without_version.get("metadata")["productVersionName"],
        )

    def test_invalid_product_mapping_key_check(self, push_base_object):
        with pytest.raises(ValueError) as exception:
            push_base_object.product_mapping.get(1)
        assert str(exception.value) == "key must be tuple (product, product_code)"

        with pytest.raises(ValueError) as exception:
            push_base_object.product_mapping.get((1,))
        assert str(exception.value) == "key must be tuple (product, product_code)"

    def test_invalid_product_version_mapping_key_check(self, push_base_object):
        with pytest.raises(ValueError) as exception:
            push_base_object.pv_mapping.get(1)
        assert str(exception.value) == "key must be tuple (product, product_code, product_version)"

        with pytest.raises(ValueError) as exception:
            push_base_object.pv_mapping.get((1,))
        assert str(exception.value) == "key must be tuple (product, product_code, product_version)"

    def test_invalid_file_mapping_key_check(self, push_base_object):
        with pytest.raises(ValueError) as exception:
            push_base_object.file_mapping.get(1)
        assert str(exception.value) == "key must be tuple (product, product_code, product_version, downloadURL)"

        with pytest.raises(ValueError) as exception:
            push_base_object.file_mapping.get((1,))
        assert str(exception.value) == "key must be tuple (product, product_code, product_version, downloadURL)"


class TestFetcherDict:
    def test_setitem(self, push_base_object):
        fetcher = FetcherDict({}, fetcher={}, key_checker=push_base_object._product_mapping_key_check)
        fetcher[("anything", "anything")] = "test"
        assert ("anything", "anything") in fetcher.data

    def test_getitem(self, push_base_object):
        fetcher = FetcherDict({}, fetcher={}, key_checker=push_base_object._product_mapping_key_check)
        fetcher[("anything", "anything")] = "test"
        result = fetcher[("anything", "anything")]
        assert "test" == result

    def test_getitem_with_empty_data(self, push_base_object):
        def fake_product(product_name, product_code):
            return {(("test_product_name", "test_product_code"), "test"): None}

        fetcher = FetcherDict({}, fetcher=fake_product, key_checker=push_base_object._product_mapping_key_check)
        fetcher[("anything", "anything")] = "test"
        result = fetcher[("test_product_name", "test_product_code")]
        assert "test" == result

    def test_delitem(self, push_base_object):
        fetcher = FetcherDict({}, fetcher={}, key_checker=push_base_object._product_mapping_key_check)
        key = ("anything", "anything")
        fetcher[key] = "test"
        del fetcher[key]
        assert key not in fetcher.data


class TestRollbackOperations:
    @pytest.fixture()
    def push_base_object(self):
        push_base = PushBase("mock://test.com/", "foo", "bar")
        return push_base

    def test_rollback_created_product(self, push_base_object):
        data = {
            "type": "product",
            "action": "create",
            "product_id": 1234,
            "metadata": {
                # some data
            },
        }
        push_base_object.completed_operations = [data]
        with requests_mock.Mocker() as m:
            m.register_uri(
                "DELETE",
                "mock://test.com/products/1234",
                status_code=200,
            )
            push_base_object.rollback_cgw_operation()
            assert m.call_count == 1

    def test_rollback_updated_product(self, push_base_object):
        data = {
            "type": "product",
            "action": "update",
            "metadata": {
                "id": 1111,
                "name": "nameTest",
                "productCode": "productCodeTest",
                "homepage": "https://test.com/",
                "downloadpage": "https://testing.com/",
                "thankYouPage": "https://testing.com/",
                "eloquaCode": "NOT_SET",
                "featuredArtifactType": "Server",
                "thankYouTimeout": 3,
            },
        }
        push_base_object.completed_operations = [data]
        push_base_object.product_records[1111] = {"id": 1111}
        with requests_mock.Mocker() as m:
            m.register_uri(
                "POST",
                "mock://test.com/products",
                status_code=200,
            )
            push_base_object.rollback_cgw_operation()
            assert m.call_count == 1

    def test_rollback_created_version(self, push_base_object):
        data = {
            "type": "product_version",
            "action": "create",
            "version_id": 2222,
            "metadata": {
                "productId": 1111,
                "versionName": "TestName",
                # some other data
            },
        }
        push_base_object.completed_operations = [data]
        with requests_mock.Mocker() as m:
            m.register_uri(
                "DELETE",
                "mock://test.com/products/1111/versions/2222",
                status_code=200,
            )
            push_base_object.rollback_cgw_operation()
            assert m.call_count == 1

    def test_rollback_update_versions(self, push_base_object):
        data = {
            "type": "product_version",
            "action": "update",
            "metadata": {
                "id": 2222,
                "productId": 1111,
                "versionName": "AnsibleNewTestVersion 1",
                "ga": True,
                "masterProductVersion": None,
                "termsAndConditions": "Anonymous Download",
                "trackingDisabled": True,
                "hidden": True,
                "invisible": False,
                "releaseDate": "2022-25-05",
            },
        }
        push_base_object.version_records[data["metadata"]["id"]] = data["metadata"]
        push_base_object.completed_operations = [data]
        with requests_mock.Mocker() as m:
            m.register_uri(
                "POST",
                "mock://test.com/products/1111/versions",
                status_code=200,
            )
            push_base_object.rollback_cgw_operation()
            assert m.call_count == 1

    def test_rollback_created_file(self, push_base_object):
        data = {
            "type": "file",
            "action": "create",
            "product_id": 1111,
            "file_id": 333,
            "metadata": {
                "productVersionId": 2222,
                "downloadURL": "/content/origin/test",
                "description": "Red Hat OpenShift Local Sandbox Test",
                "label": "Checksum File Sandbox Test label",
                # some other data
            },
        }
        push_base_object.completed_operations = [data]
        with requests_mock.Mocker() as m:
            m.register_uri(
                "DELETE",
                "mock://test.com/products/1111/versions/2222/files/333",
                status_code=200,
            )
            push_base_object.rollback_cgw_operation()
            assert m.call_count == 1

    def test_rollback_update_files(self, push_base_object):
        data = {
            "type": "file",
            "action": "update",
            "product_id": 1111,
            "metadata": {
                "description": "Red Hat OpenShift Local Sandbox Test",
                "label": "Checksum File Sandbox Test label",
                "order": 0,
                "hidden": False,
                "invisible": False,
                "type": "FILE",
                "differentProductThankYouPage": None,
                "downloadURL": "/content/origin/test",
                "shortURL": "/pub-1/openshift-v4/testing",
                "md5": None,
                "size": None,
                "productVersionId": 2222,
                "id": 3333,
            },
        }
        push_base_object.file_records[data["metadata"]["id"]] = data["metadata"]
        push_base_object.completed_operations = [data]
        with requests_mock.Mocker() as m:
            m.register_uri(
                "POST",
                "mock://test.com/products/1111/versions/2222/files",
                status_code=200,
            )
            push_base_object.rollback_cgw_operation()
            assert m.call_count == 1


class TestMakeVisible:
    @pytest.fixture()
    def push_base_object(self):
        push_base = PushBase("mock://test.com/", "foo", "bar")
        return push_base

    def test_make_visible_created_version(self, push_base_object):
        data = {
            "type": "product_version",
            "action": "create",
            "version_id": 2222,
            "metadata": {
                "productId": 1111,
                "versionName": "AnsibleNewTestVersion 1",
                "ga": True,
                "masterProductVersion": None,
                "termsAndConditions": "Anonymous Download",
                "trackingDisabled": True,
                "hidden": True,
                "invisible": False,
                "releaseDate": "2022-25-05",
            },
        }
        push_base_object.completed_operations = [data]
        with requests_mock.Mocker() as m:
            m.register_uri(
                "POST",
                "mock://test.com/products/1111/versions",
                status_code=200,
            )
            push_base_object.make_visible()
            assert m.call_count == 1

    def test_make_visible_created_file(self, push_base_object):
        data = {
            "type": "file",
            "action": "create",
            "product_id": 1111,
            "file_id": 333,
            "metadata": {
                "hidden": False,
                "invisible": False,
                "type": "FILE",
                "downloadURL": "/content/origin/test",
                "shortURL": "/pub-1/openshift-v4/testing",
                "productVersionId": 2222,
            },
        }
        push_base_object.completed_operations = [data]
        with requests_mock.Mocker() as m:
            m.register_uri(
                "POST",
                "mock://test.com/products/1111/versions/2222/files",
                status_code=200,
            )
            push_base_object.make_visible()
            assert m.call_count == 1



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath("../pubtools/_content_gateway/"))

project = 'pubtools-content-gateway'
copyright = '2022, Red Hat'
author = 'Red Hat'
release = '0.2.0'
release = '0.2.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.githubpages",
    "sphinx.ext.viewcode",
    "sphinxarg.ext",
]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#

html_context = {
    'display_github': True,
    "github_user": "release-engineering",
    "github_repo": "pubtools-content-gateway",
    'github_version': 'main/docs/',
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



from .cgw_client import CGWClient
from .cgw_authentication import CGWBasicAuth
import logging

LOG = logging.getLogger("pubtools.cgw")


class CGWError(Exception):
    """Custom exception class to handle Content Gateway errors."""

    def __init__(self, msg):
        super(CGWError, self).__init__(msg)


class FetcherDict:
    """A class to fetch the ids of the products, versions and file
    from their mapping object
    """

    def __init__(self, val=None, fetcher=None, key_checker=None):
        """
        Initializing the FetcherDict
        The fetcher takes a fetch method as an argument and returns response.
        The key checker takes a mapping method as an argument and validate its keys.

        Args:
            fetcher (method):
                A fetch method instance
            key_checker (method):
                A mapping method instance
        """

        self.fetcher = fetcher
        self.key_checker = key_checker
        self.data = val or {}

    def __setitem__(self, key, val):
        """
        Set the data value of the specified key

        Args:
            key (tuple):
                Tuple key to set specified value
            val (value):
                Value to be set
        """
        self.key_checker(key)
        self.data[key] = val

    def __getitem__(self, key):
        """
        Returns the data value of the specified key

        Args:
            key (tuple):
                Tuple key to return value

        Returns:
            key (value):
                Returns the value of specified key.
        """

        self.key_checker(key)
        if key not in self.data and self.fetcher:
            for new_key, vid in self.fetcher(*key):
                self.data[new_key] = vid
        return self.data[key]

    def get(self, key, default=None):
        """
        Returns the data value of the specified key

        Args:
            key (tuple):
                Tuple key to return value

        Returns:
            key (value):
                Returns the value of specified key.
        """
        self.key_checker(key)
        if key not in self.data and self.fetcher:
            for new_key, vid in self.fetcher(*key):
                self.data[new_key] = vid
        return self.data.get(key, default)

    def __delitem__(self, key):
        """
        Delete the data value of the specified index

        Args:
            key (tuple):
                Tuple key to delete specified record
        """
        self.key_checker(key)
        self.data.__delitem__(key)


class PushBase:
    """Base class for all operations on content gateway via
    either through entry points or from staged input.

    Adds arguments and environment variables common to all operations.
    """

    def __init__(self, cgw_hostname, cgw_username, cgw_password):
        """
        Initialize.

        Args:
            cgw_hostname (str):
                CGW registry URL
            cgw_username (str):
                username for CGW HTTP API
            cgw_password (str):
                password for CGW HTTP API
        """

        self.auth = CGWBasicAuth(cgw_username, cgw_password)
        self.cgw_client = CGWClient(cgw_hostname, self.auth)
        self.product_mapping = FetcherDict({}, fetcher=self._fetch_product, key_checker=self._product_mapping_key_check)
        self.pv_mapping = FetcherDict(
            {},
            fetcher=self._fetch_product_version,
            key_checker=self._product_version_mapping_key_check,
        )
        self.file_mapping = FetcherDict({}, fetcher=self._fetch_file, key_checker=self._file_key_check)
        self.product_records = {}
        self.version_records = {}
        self.file_records = {}
        self.completed_operations = []

    @staticmethod
    def _product_mapping_key_check(key):
        """
        Checks the product mapping keys. The keys must follow the following constraints:
            1) The key must be of tuple type
            2) Two keys must be present in the key tuple

        Args:
            key (tuple(str, str)):
                The keys should contain product, product_code

        Raises:
            ValueError:
                When the keys are not of tuple type or
                the keys in tuple doesn't satisfy the length size.
        """

        if not isinstance(key, tuple):
            raise ValueError("key must be tuple (product, product_code)")
        elif not len(key) == 2:
            raise ValueError("key must be tuple (product, product_code)")

    @staticmethod
    def _product_version_mapping_key_check(key):
        """
        Checks the product version mapping keys. The keys must follow the following constraints:
            1) The key must be of tuple type
            2) Three keys must be present in the key tuple

        Args:
            key (tuple(str, str, str|int)):
                The keys should contain product, product_code and product_version

        Raises:
            ValueError:
                When the keys are not of tuple type or
                the keys in tuple doesn't satisfy the length size.
        """

        if not isinstance(key, tuple):
            raise ValueError("key must be tuple (product, product_code, product_version)")
        elif not len(key) == 3:
            raise ValueError("key must be tuple (product, product_code, product_version)")

    @staticmethod
    def _file_key_check(key):
        """
        Checks the product version mapping keys. The keys must follow the following constraints:
            1) The key must be of tuple type
            2) Four keys must be present in the key tuple

        Args:
            key (tuple(str, str, str|int, str)):
                The keys should contain product, product_code,
                 product_version and downloadURL

        Raises:
            ValueError:
                When the keys are not of tuple type or
                the keys in tuple doesn't satisfy the length size.
        """

        if not isinstance(key, tuple):
            raise ValueError("key must be tuple (product, product_code, product_version, downloadURL)")
        elif not len(key) == 4:
            raise ValueError("key must be tuple (product, product_code, product_version, downloadURL)")

    def _fetch_product(self, product_name, product_code):
        """
        Get the list of products from the content gateway
        by calling its client methods.

        Args:
            product_name (str):
                product name of the product
            product_code (str):
                product code of the product

        Returns:
            Yields (tuple):
                Returns the product record. The returned structure
                is an iterator.

                Intended to be called as the first operation in
                the content gateway workflow.
        """

        all_products = self.cgw_client.get_products()
        for product in all_products:
            self.product_records[product["id"]] = product
            yield (product["name"], product["productCode"]), product["id"]

    def _fetch_product_version(self, product_name, product_code, version_name):
        """
        Get the list of product versions from the content gateway
        by calling its client methods.

        Args:
            product_name (str):
                product name of the product
            product_code (str):
                product code of the product
            version_name (str|int):
                version name of the product

        Returns:
            Yields (tuple):
                Returns the version records associate with the product.
                The returned structure is an iterator.
        """

        product_id = self.product_mapping.get((product_name, product_code))
        all_versions = self.cgw_client.get_versions(product_id)
        for version in all_versions:
            self.version_records[version["id"]] = version
            yield (product_name, product_code, version["versionName"]), version["id"]

    def _fetch_file(self, product_name, product_code, version_name, download_url):
        """
        Get the list of file records associate with version and product
        from the content gateway by calling its client methods.

        Args:
            product_name (str):
                product name of the file
            product_code (str):
                product code of the file
            version_name (str|int):
                version name of the file
            download_url (str):
                download url of the file

        Returns:
            Yields (dict):
                Returns the file records.
                The returned structure is an iterator.
        """

        product_id = self.product_mapping.get((product_name, product_code))
        version_id = self.pv_mapping.get((product_name, product_code, version_name))
        all_files = self.cgw_client.get_files(product_id, version_id)
        for data in all_files:
            self.file_records[data["id"]] = data
            yield (
                product_name,
                product_code,
                version_name,
                data["downloadURL"],
            ), data.get("id")

    def process_product(self, item):
        """
        Responsible for product workflow operations such as Create,
        Update and delete the product based on the satisfied condition.

        The product operation follows the following workflow:
            - Search for the product record from the product mapping
            - If `product action` == `create`

                - Creates a new product record on content gateway with given data
                - Add the new record into the product mapping records

            - If `product action` == `update` and record found in mapping or id is provided in metadata

                - Update the product record on content gateway with given data
                - Update the record into the product mapping records

            - If `product action` == `delete` and record found in mapping or id is provided in metadata

                - Delete the product record on the content gateway
                - Remove the record from the product mapping records

            - If `product action` == `delete` or `update` and record not found in mapping

                - Raises CGWError error

        Args:
            item (dict):
                Metadata of product record

        Raises:
            CGWError:
                When product record is not present and tries to update/delete the same

        Returns:
            product_id (int):
                Returns the product ID
        """

        product_name = item.get("metadata").get("name")
        product_code = item.get("metadata").get("productCode")
        LOG.debug(
            "Fetching for product_id of product_name:- %s and product_code:- %s \n" % (product_name, product_code)
        )

        #  Trying to fetch product_id from CGW mapping.
        product_id = self.product_mapping.get((product_name, product_code))

        #  If CGW mapping doesn't have the record it will try to get product_id from metadata if provided
        #  For the create operation it will not get any product_id either from the CGW
        #  or from metadata. The product_id would be None
        #  To update the product name and productCode the product_id must be provided in the metadata
        product_id = product_id if product_id else item.get("metadata").get("id")

        if item.get("action") == "create":
            if product_id:
                LOG.error(
                    "Record already present for the product name:- %s and code:- %s and product_id is %s "
                    % (product_name, product_code, product_id)
                )
                raise CGWError("Cannot create new product. Record already present.")

            product_id = self.cgw_client.create_product(item.get("metadata"))
            # A new product is created so updating the existing product mapping
            self.product_mapping[(product_name, product_code)] = product_id
            LOG.info("Created a new product with product_id:- %s\n" % product_id)

            # adding operational product_id for rollback operation
            item["product_id"] = product_id
            # keeping a copy of data that has been used to create the cgw data
            self.completed_operations.append(item)
            return product_id

        elif item.get("action") == "update" and product_id:
            item.get("metadata")["id"] = product_id
            LOG.info("Updating product record of id:- %s \n", product_id)
            self.cgw_client.update_product(item.get("metadata"))

            # keeping a copy of data that has been used to update the cgw data
            self.completed_operations.append(item)

        elif item.get("action") == "delete" and product_id:
            LOG.info("Deleting existing product records for product_id:- %s" % product_id)
            self.cgw_client.delete_product(product_id)
            del self.product_mapping[(product_name, product_code)]

            #  removing the deleted file record from the completed operations
            #  this is to avoid the conflicts during the rollback and make_visible operations
            for record in self.completed_operations:
                if record["type"] == "product" and record["product_id"] == product_id:
                    self.completed_operations.remove(record)

            LOG.info("Product record deleted! \n")
        else:
            raise CGWError("Cannot update/delete the product %s %s, id is not set" % (product_name, product_code))
        return product_id

    def process_version(self, item):
        """
        Responsible for version workflow operations such as Create,
        Update and delete the product version based on the satisfied condition
        on content gateway (CGW).

        The product version operation follows the following workflow:
            - Search for the product record from the product mapping
            - Search for the product version record from the product_version(pv_mapping) mapping
            - If `version action` == `create`

                - Creates a new version record associated with product on CGW with given data
                - Add the new record into the mapping records

            - If `version action` == `update` and record found in mapping or id is provided in metadata

                - Update the version record on content gateway with given data
                - Update the record into the version mapping records

            - If `version action` == `delete` and record found in mapping or id is provided in metadata

                - Delete the version record on the content gateway
                - Remove the record from the version mapping records

            - If `version action` == `delete` or `update` and record not found in mapping
               or id is not provided in metadata

                - Raises CGWError error

        Args:
            item (dict):
                Metadata of version record

        Raises:
            CGWError:
                - When product record is not present
                - When version record is not present and tries to update/delete the same

        Returns:
            version_id (int|None):
                Returns the version ID
        """

        product_name = item.get("metadata")["productName"]
        product_code = item.get("metadata")["productCode"]
        version_name = item.get("metadata")["versionName"]

        product_id = self.product_mapping.get((product_name, product_code))
        if not product_id:
            raise CGWError("Product name: %s product code: %s not found" % (product_name, product_code))

        LOG.debug(
            "Fetching the product and version_id of "
            "product_name:- %s and version_name:- %s \n" % (product_name, version_name)
        )

        #  Trying to fetch version_id from CGW mapping.
        version_id = self.pv_mapping.get((product_name, product_code, version_name))

        #  If CGW mapping doesn't have the record it will try to get version_id from metadata if provided
        #  For the create operation it will not get any version_id either from the CGW
        #  or from metadata. The version_id would be None
        #  To update the versionName the version_id must be provided in the metadata
        version_id = version_id if version_id else item.get("metadata").get("id")

        item.get("metadata")["productId"] = product_id
        del item.get("metadata")["productName"], item.get("metadata")["productCode"]

        if item.get("action") == "create":
            if version_id:
                LOG.error(
                    "Record already present for the version name:- %s and code:- %s and version_id is %s "
                    % (version_name, product_code, version_id)
                )
                raise CGWError("Cannot create new version. Record already present.")
            LOG.info("Creating version entry for the given version_name: %s ", version_name)
            version_id = self.cgw_client.create_version(product_id, item.get("metadata"))
            # adding new version record to the version mapping
            self.pv_mapping[(product_name, product_code, version_name)] = version_id
            LOG.info("New version created with version_id:- %s \n" % version_id)

            # adding operational version_id for rollback operation
            item["version_id"] = version_id
            # keeping a copy of data that has been used to create the cgw data
            self.completed_operations.append(item)
            return

        if item.get("action") == "update" and version_id:
            item.get("metadata")["id"] = version_id
            LOG.info("Updating the version metadata \n")
            self.cgw_client.update_version(product_id, item.get("metadata"))
            # keeping a copy of data that has been used to update the cgw data
            self.completed_operations.append(item)

        elif item.get("action") == "delete" and version_id:
            LOG.info("Deleting existing version records for version_id:- %s" % version_id)
            self.cgw_client.delete_version(product_id, version_id)
            del self.pv_mapping[(product_name, product_code, version_name)]

            #  removing the deleted file record from the completed operations
            #  this is to avoid the conflicts during the rollback and make_visible operations
            for record in self.completed_operations:
                if record["type"] == "product_version" and record["version_id"] == version_id:
                    self.completed_operations.remove(record)

            LOG.info("Version record deleted! \n")
        else:
            raise CGWError(
                "Cannot update/delete the version name: '%s' product code: '%s' product: '%s', id is not set"
                % (version_name, product_code, product_name)
            )
        return version_id

    def process_file(self, file_item):
        """
        Responsible for file workflow operations such as Create,
        Update and delete the file based on the satisfied condition
        on content gateway (CGW).

        The file operation follows the following workflow:
            - Search for the product record from the product mapping
            - Search for the product version record from the product_version (pv_mapping) mapping
            - Search for the file record from the file mapping (file_mapping) mapping
            - If `file action` == `create`

                - Creates a new file record associated with product on CGW with given data
                - Add the new record into the file mapping records

            - If `file action` == `update` and record found in mapping or id is provided in metadata

                - Update the file record on content gateway with given data
                - Update the record into the file mapping records

            - If `file action` == `delete` and record found in mapping or id is provided in metadata

                - Delete the file record on the content gateway
                - Remove the record from the file mapping records

            - If `file action` == `delete` or `update` and record not found in mapping
              or id is not provided in metadata

                - Raises CGWError error

            - The default file `order` value will be `10` if provided none
            - file `order` will be auto incremented by `10` for nested data yml structure

        Args:
            file_item (dict):
                Metadata of file record

        Raises:
            CGWError:
                - When product record is not present
                - When version record is not present
                - When file record is not present and tries to update/delete the same

        Returns:
            file_id (int|None):
                Returns the file ID
        """

        LOG.debug("Fetching file id for the file metadata \n")
        product_name = file_item.get("metadata").get("productName")
        product_code = file_item.get("metadata").get("productCode")
        version_name = file_item.get("metadata").get("productVersionName")
        download_url = file_item.get("metadata").get("downloadURL")
        product_id = self.product_mapping.get((product_name, product_code))

        if not product_id:
            raise CGWError("Product name: %s product code: %s not found" % (product_name, product_code))

        version_id = self.pv_mapping.get((product_name, product_code, version_name))
        if not version_id:
            raise CGWError(
                "Product name: %s product code: %s version name: %s not found"
                % (product_name, product_code, version_name)
            )

        #  Trying to fetch file_id from CGW mapping.
        file_id = self.file_mapping.get((product_name, product_code, version_name, download_url))

        #  If CGW mapping doesn't have the record it will try to get file_id from metadata if provided
        #  For the create operation it will not get any file_id neither from the CGW
        #  nor from metadata. The file_id would be None
        #  To update the downloadURL the file_id must be provided in the metadata
        file_id = file_id if file_id else file_item.get("metadata").get("id")

        del (
            file_item.get("metadata")["productName"],
            file_item.get("metadata")["productCode"],
            file_item.get("metadata")["productVersionName"],
        )
        file_item.get("metadata")["productVersionId"] = version_id

        if file_item.get("action") == "create":
            if file_id:
                LOG.error(
                    "Record already present for the file download_url :- %s and file_id is %s "
                    % (download_url, file_id)
                )
                raise CGWError("Cannot create new file. Record already present.")
            LOG.info("Creating file entry for the given downloadURL: %s ", download_url)
            file_metadata = file_item.get("metadata")
            if "pushItemPath" in file_metadata:
                file_metadata.pop("pushItemPath")
            file_id = self.cgw_client.create_file(product_id, version_id, file_metadata)
            self.file_mapping[(product_name, product_code, version_name, download_url)] = file_id
            LOG.info("New file created with file_id:- %s \n" % file_id)

            # adding operational product_id and file_id for rollback operation
            file_item["product_id"] = product_id
            file_item["file_id"] = file_id
            # keeping a copy of data that has been used to create the cgw data
            self.completed_operations.append(file_item)
            return

        if file_item.get("action") == "update" and file_id:
            LOG.info("Updating the file record \n")
            file_item.get("metadata")["id"] = file_id
            self.cgw_client.update_file(product_id, version_id, file_item.get("metadata"))

            # adding operational product_id for rollback operation
            file_item["product_id"] = product_id
            # keeping a copy of data that has been used to update the cgw data
            self.completed_operations.append(file_item)

        elif file_item.get("action") == "delete" and file_id:
            LOG.info("Deleting existing file records for file_id:- %s" % file_id)
            self.cgw_client.delete_file(product_id, version_id, file_id)

            #  removing the deleted file record from the completed operations
            #  this is to avoid the conflicts during the rollback and make_visible operations
            for record in self.completed_operations:
                if record["type"] == "file" and record["file_id"] == file_id:
                    self.completed_operations.remove(record)

            LOG.info("File record deleted! \n")
            del self.file_mapping[(product_name, product_code, version_name, download_url)]
        else:
            raise CGWError(
                "Cannot update/delete file: product version: '%s' product code: '%s' product: '%s', id is not set"
                % (version_name, product_code, product_name)
            )
        return file_id

    def make_visible(self):
        """

        - Update the invisible attribute to False.
          once all the CGW operation completed successfully.

        - This is needed to make the version and file records visible.
        - The `invisible` attribute changes on the following conditions:

            - `invisible=True` :- sets if user externally specify the `invisible=True`
            - `invisible=False` :- sets either if user externally specify the `invisible=False`
              or user doesn't specify any value for this `invisible` attribute.

        """

        LOG.debug("Updating the invisible attribute of the created records")
        LOG.debug("Enabling the created records if required \n")

        self.completed_operations.reverse()
        for item in self.completed_operations:
            if item.get("type") == "product_version":
                if item.get("action") == "create":
                    item["metadata"]["id"] = item["version_id"]
                    item["metadata"]["invisible"] = False if not item["metadata"].get("invisible") else True
                    self.cgw_client.update_version(item["metadata"]["productId"], item["metadata"])

            elif item.get("type") == "file":
                if item.get("action") == "create":
                    item["metadata"]["id"] = item["file_id"]
                    item["metadata"]["invisible"] = False if not item["metadata"].get("invisible") else True
                    self.cgw_client.update_file(
                        item["product_id"],
                        item["metadata"]["productVersionId"],
                        item["metadata"],
                    )

    def rollback_cgw_operation(self):
        """

        Undo the partial completed CGW operation.
            - For the create operations it will delete the created data from CGW
            - For the update operation it will revert the changes with old data.

        """

        self.completed_operations.reverse()
        for item in self.completed_operations:
            if item.get("type") == "product":
                if item.get("action") == "create":
                    self.cgw_client.delete_product(item["product_id"])
                    LOG.info("Rollback done for created product:- %s", item["product_id"])

                elif item.get("action") == "update":
                    item["metadata"] = self.product_records.get(item["metadata"]["id"])
                    self.cgw_client.update_product(item["metadata"])
                    LOG.info("Rollback done for updated product:- %s", item["metadata"]["id"])

            elif item.get("type") == "product_version":
                if item.get("action") == "create":
                    self.cgw_client.delete_version(item["metadata"]["productId"], item["version_id"])
                    LOG.info(
                        "Rollback done for created product_version:- %s",
                        item["version_id"],
                    )

                elif item.get("action") == "update":
                    item["metadata"] = self.version_records.get(item["metadata"]["id"])
                    self.cgw_client.update_version(item["metadata"]["productId"], item["metadata"])
                    LOG.info(
                        "Rollback done for updated product_version:- %s",
                        item["metadata"]["id"],
                    )

            elif item.get("type") == "file":
                if item.get("action") == "create":
                    self.cgw_client.delete_file(
                        item["product_id"],
                        item["metadata"]["productVersionId"],
                        item["file_id"],
                    )
                    LOG.info("Rollback done for created file:- %s", item["file_id"])

                elif item.get("action") == "update":
                    item["metadata"] = self.file_records.get(item["metadata"]["id"])
                    self.cgw_client.update_file(
                        item["product_id"],
                        item["metadata"]["productVersionId"],
                        item["metadata"],
                    )
                    LOG.info("Rollback done for updated file:- %s", item["metadata"]["id"])
        LOG.info("Rollback operations completed...!")



import os
import argparse
import pytest
from pubtools._content_gateway.push_cgw import PushCGW, main
from tests.fake_cgw_client import TestClient

try:
    import mock
except ImportError:
    from unittest import mock

yaml_file_path = os.path.join(os.path.dirname(__file__), "test_data/test_cgw_push.yaml")


@mock.patch("pubtools._content_gateway.push_base.CGWClient", return_value=TestClient())
def test_cgw_operations(mocked_cgw_client):
    push_cgw = PushCGW("http://fake_host_name/test", "foo", "bar", yaml_file_path)
    push_cgw.cgw_operations()

    assert len(push_cgw.cgw_client.create_product.calls) >= 2
    assert len(push_cgw.cgw_client.update_product.calls) >= 1
    assert len(push_cgw.cgw_client.delete_product.calls) >= 1
    assert len(push_cgw.cgw_client.create_version.calls) >= 2
    assert len(push_cgw.cgw_client.update_version.calls) >= 1
    assert len(push_cgw.cgw_client.delete_version.calls) >= 1
    assert len(push_cgw.cgw_client.create_file.calls) >= 1
    assert len(push_cgw.cgw_client.delete_file.calls) >= 1
    assert mocked_cgw_client.called is True


@mock.patch("pubtools._content_gateway.push_cgw.yaml_parser")
@mock.patch("pubtools._content_gateway.push_base.CGWClient", return_value=TestClient())
@mock.patch(
    "pubtools._content_gateway.push_cgw.argparse.ArgumentParser.parse_args",
    return_value=argparse.Namespace(
        CGW_hostname="http://fake_host_name/test",
        CGW_username="test_username",
        CGW_password="**********",
        CGW_filepath=yaml_file_path,
    ),
)
def test_main(mock_args, mock_cgw, mock_yaml_parser, create_product_data):
    mock_yaml_parser.return_value = [create_product_data]
    main()
    expected_product_id = 1
    assert expected_product_id in mock_cgw.return_value.products
    assert mock_args.called is True


def test_cgw_operations_exception():
    push_cgw = PushCGW("http://fake_host_name/test", "foo", "bar", yaml_file_path)
    push_cgw.process_product = []
    push_cgw.rollback_cgw_operation = mock.MagicMock()
    with pytest.raises(TypeError):
        push_cgw.cgw_operations()

    assert push_cgw.rollback_cgw_operation.called is True


@pytest.fixture
def mock_cgw_password():
    os.environ["CGW_PASSWORD"] = "test_password_from_env"
    yield
    del os.environ["CGW_PASSWORD"]


@mock.patch("pubtools._content_gateway.push_cgw.PushCGW")
@mock.patch(
    "pubtools._content_gateway.push_cgw.argparse.ArgumentParser.parse_args",
    return_value=argparse.Namespace(
        CGW_hostname="http://fake_host_name/test",
        CGW_username="test_username",
        CGW_password=None,
        CGW_filepath=yaml_file_path,
    ),
)
def test_main_with_env_var_password(mock_args, mock_push_cgw, mock_cgw_password):

    main()

    mock_push_cgw.assert_called_once_with(
        "http://fake_host_name/test", "test_username", "test_password_from_env", yaml_file_path
    )

    assert mock_args.called is True


@mock.patch("pubtools._content_gateway.push_cgw.PushCGW")
@mock.patch(
    "pubtools._content_gateway.push_cgw.argparse.ArgumentParser.parse_args",
    return_value=argparse.Namespace(
        CGW_hostname="http://fake_host_name/test",
        CGW_username="test_username",
        CGW_password=None,
        CGW_filepath=yaml_file_path,
    ),
)
def test_main_fail_without_password(mock_args, mock_push_cgw, capfd):
    if "CGW_PASSWORD" in os.environ:
        del os.environ["CGW_PASSWORD"]

    with pytest.raises(SystemExit) as excinfo:
        try:
            main()
        except SystemExit:
            captured = capfd.readouterr()
            assert "CGW password must be provided" in captured.err
            raise

    assert excinfo.value.code == 2

    mock_push_cgw.assert_not_called()

    assert mock_args.called is True



import argparse
import logging
import os
from .push_base import PushBase
from .utils import yaml_parser, validate_data, sort_items, format_cgw_items

LOG = logging.getLogger("pubtools.cgw")
LOG_FORMAT = "%(asctime)s [%(levelname)-8s] %(message)s"


class PushCGW(PushBase):
    """Handle push CGW workflow."""

    def __init__(self, cgw_hostname, cgw_username, cgw_password, cgw_filepath):
        """
        Initialize.

        Args:
            cgw_hostname (str):
                CGW registry URL
            cgw_username (str):
                username for CGW HTTP API
            cgw_password (str):
                password for CGW HTTP API
            cgw_filepath (str):
                filepath of the yaml file
        """

        PushBase.__init__(self, cgw_hostname, cgw_username, cgw_password)
        self.cgw_filepath = cgw_filepath
        self.cgw_items = []

    def cgw_operations(self):
        """
        Initiate the CGW operations such as create, update or delete
        on products, versions and files.
        """

        self.cgw_items = yaml_parser(self.cgw_filepath)
        self.cgw_items = format_cgw_items(self.cgw_items)

        for item in self.cgw_items:
            validate_data(item)
        self.cgw_items = sort_items(self.cgw_items)
        try:
            for item in self.cgw_items:
                if item["type"] == "product":
                    self.process_product(item)
                elif item["type"] == "product_version":
                    self.process_version(item)
                elif item["type"] == "file":
                    self.process_file(item)

            self.make_visible()
            LOG.info("\n All CGW operations are successfully completed...!")
        except Exception as error:
            LOG.exception("Exception occurred during the CGW operation %s" % error)
            LOG.info("Rolling back the partial operation")
            self.rollback_cgw_operation()

            #  raising the occurred Exception as all the exception will get caught in this except block
            #  we want to return full Traceback
            raise error


def main():
    """Entrypoint for CGW Push."""

    logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
    parser = argparse.ArgumentParser()
    parser.add_argument("-host", "--CGW_hostname", required=True, metavar="CGW-hostname", help="Hostname of the server")
    parser.add_argument(
        "-u", "--CGW_username", required=True, metavar="CGW-username", help="Username of Content Gateway"
    )
    parser.add_argument(
        "-p",
        "--CGW_password",
        metavar="CGW-password",
        help="Password for Content Gateway (optional if set as \
        environment variable)",
    )
    parser.add_argument(
        "-f", "--CGW_filepath", required=True, metavar="CGW-filepath", help="File path to read metadata"
    )
    args = parser.parse_args()

    # Check if password is provided as an argument or through an environment variable
    cgw_password = args.CGW_password or os.getenv("CGW_PASSWORD")

    # If neither is provided, raise an error
    if not cgw_password:
        parser.error("CGW password must be provided as an argument or set as the 'CGW_PASSWORD' environment variable.")

    push_cgw = PushCGW(args.CGW_hostname, args.CGW_username, cgw_password, args.CGW_filepath)
    push_cgw.cgw_operations()



import os
from pubtools._content_gateway.utils import yaml_parser, validate_data, sort_items, format_cgw_items

test_data_dir = os.path.join(os.path.dirname(__file__), "test_data")


def test_validate_data(fixture_sorted_files_json):
    for data in fixture_sorted_files_json:
        assert validate_data(data) is True


def test_yaml_parser():
    yaml_file = os.path.join(test_data_dir, "test_cgw_push.yaml")
    yaml_parser(yaml_file)


def test_sort_items(fixture_unsorted_files_json, fixture_sorted_files_json):
    assert fixture_sorted_files_json == sort_items(fixture_unsorted_files_json)


def test_yml_data_formats(yml_json_data):
    yaml_file = os.path.join(test_data_dir, "test_yml_format.yaml")
    cgw_items = yaml_parser(yaml_file)
    formatted_cgw_data = format_cgw_items(cgw_items)
    assert formatted_cgw_data == yml_json_data



import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


class CGWSession(object):
    """Helper class to support cgw requests and authentication."""

    def __init__(self, hostname, retries=3, verify=True, backoff_factor=2):
        """Initializing.

        Args:
            hostname (str)
                hostname of CG service
            retries (int)
                number of http retries
            verify (bool)
                enable/disable SSL verification
            backoff_factor (int)
                backoff factor to apply between attempts after the second try
        """
        self.session = requests.Session()
        self.hostname = hostname
        self.verify = verify

        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=set(range(500, 512)),
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)
        self.session.headers["Content-type"] = "application/json"
        self.session.headers["Accept"] = "application/json"

    def get(self, endpoint, **kwargs):
        """HTTP get request against CGW server API

        Args:
            endpoint (str)
                API specific endpoint for the request
        Returns:
            requests.Response
        """

        return self.session.get(self._api_url(endpoint), verify=self.verify, **kwargs)

    def post(self, endpoint, **kwargs):
        """HTTP post request against CGW server API.

        Args:
            endpoint (str)
                API specific endpoint for the request
        Returns:
            requests.Response
        """

        return self.session.post(self._api_url(endpoint), verify=self.verify, **kwargs)

    def put(self, endpoint, **kwargs):
        """HTTP put request against CGW server API.

        Args:
            endpoint (str)
                API specific endpoint for the request
        Returns:
            requests.Response
        """

        return self.session.put(self._api_url(endpoint), verify=self.verify, **kwargs)

    def delete(self, endpoint, **kwargs):
        """HTTP delete request against CGW server API.

        Args:
            endpoint (str)
                API specific endpoint for the request
        Returns:
            requests.Response
        """

        return self.session.delete(self._api_url(endpoint), verify=self.verify, **kwargs)

    def _api_url(self, endpoint):
        """Basic authentication support for CGClient.

        Args:
            endpoint (str)
                API specific endpoint for the request
        Returns:
            requests.Response
        """

        return "%s%s" % (self.hostname, endpoint)



from mock import MagicMock
from pubtools._content_gateway.cgw_authentication import (
    CGWAuth,
    CGWBasicAuth,
    CGWClientError,
)
from pubtools._content_gateway.cgw_client import CGWClient
from base64 import b64encode
import pytest


def test_client_auth():
    auth = CGWBasicAuth("foo", "bar")
    cgw_client = CGWClient("fake-host", cgw_auth=auth)
    user_pass_string = "%s:%s" % ("foo", "bar")
    user_pass = b64encode(user_pass_string.encode("utf-8")).decode("ascii")
    assert cgw_client.cgw_session.session.headers["Authorization"] == "Basic %s" % user_pass


def test_cgw_basic_auth():
    session = MagicMock()
    session.session.headers = {}
    auth = CGWBasicAuth("foo", "bar")
    auth.make_auth(session)

    user_pass_string = "%s:%s" % ("foo", "bar")
    user_pass = b64encode(user_pass_string.encode("utf-8")).decode("ascii")
    assert session.session.headers["Authorization"] == "Basic %s" % user_pass


def test_invalid_cgw_auth():
    session = MagicMock()
    session.session.headers = {}
    auth = CGWBasicAuth("", "")
    with pytest.raises(CGWClientError) as exception:
        auth.make_auth(session)
    assert "Error: username / password not found !!" in str(exception.value.message)


def test_cgwauth_abstract():
    try:
        CGWAuth()
        raise AssertionError("Should raise NotImplementedError")
    except NotImplementedError:
        pass



import pytest
import requests_mock
from pubtools._content_gateway.cgw_client import (
    CGWClient,
    CGWClientError,
)
from pubtools._content_gateway.cgw_authentication import CGWBasicAuth


@pytest.fixture
def get_all_products_response_json():
    json = [
        {
            "id": 111111,
            "name": "Cloud: Ansible Test",
            "productCode": "AnsibleTest",
            "latestURL": "null",
            "created": 1111111111111,
            "updated": 1111111111111,
            "modifiedBy": "R: test@example.com",
            "homepage": "https://test.fixture.com/",
            "downloadpage": "https://test.fixture.com/",
            "thankYouPage": "https://test.fixture.com/",
            "eloquaCode": "NOT_SET",
            "featuredArtifactType": "Server",
            "thankYouTimeout": 5,
        },
        {
            "id": 222222,
            "name": "Openstack: Server Test",
            "productCode": "OpenstackServerTest",
            "latestURL": "null",
            "created": 22222222,
            "updated": 22222222,
            "modifiedBy": "R: test@example.com",
            "homepage": "https://test.fixture.com/",
            "downloadpage": "https://test.fixture.com/",
            "thankYouPage": "https://test.fixture.com/",
            "eloquaCode": "NOT_SET",
            "featuredArtifactType": "Server",
            "thankYouTimeout": 5,
        },
    ]
    return json


@pytest.fixture
def product_response_json():
    json = {
        "id": 111111,
        "name": "Cloud: Ansible NEW Test",
        "productCode": "AnsibleNewTest",
        "latestURL": "null",
        "created": 1111111111111,
        "updated": 1111111111111,
        "modifiedBy": "R: test@example.com",
        "homepage": "https://test.fixture.com/",
        "downloadpage": "https://test.fixture.com/",
        "thankYouPage": "https://test.fixture.com/",
        "eloquaCode": "NOT_SET",
        "featuredArtifactType": "Server",
        "thankYouTimeout": 5,
    }
    return json


@pytest.fixture
def get_all_versions_response_json():
    json = [
        {
            "id": 333333,
            "productId": 111111,
            "versionName": "v1",
            "ga": False,
            "created": 1619209855705,
            "updated": 1619209855705,
            "modifiedBy": "R: jalam",
            "masterProductVersion": None,
            "termsAndConditions": "Anonymous Download",
            "trackingDisabled": False,
            "hidden": False,
            "invisible": False,
            "releaseDate": 1619136000000,
        },
        {
            "id": 444444,
            "productId": 111111,
            "versionName": "v2",
            "ga": False,
            "created": 1619209855705,
            "updated": 1619209855705,
            "modifiedBy": "R: jalam",
            "masterProductVersion": None,
            "termsAndConditions": "Anonymous Download",
            "trackingDisabled": False,
            "hidden": False,
            "invisible": False,
            "releaseDate": 1619136000000,
        },
    ]
    return json


@pytest.fixture
def version_response_json():
    json = {
        "id": 333333,
        "productId": 111111,
        "versionName": "v1",
        "ga": False,
        "created": 1619209855705,
        "updated": 1619209855705,
        "modifiedBy": "R: jalam",
        "masterProductVersion": None,
        "termsAndConditions": "Anonymous Download",
        "trackingDisabled": False,
        "hidden": False,
        "invisible": False,
        "releaseDate": 1619136000000,
    }
    return json


@pytest.fixture
def get_all_files_response_json():
    json = [
        {
            "id": 111111,
            "description": "Red Hat OpenShift Local",
            "label": "Checksum File",
            "order": 0,
            "created": 1111111111111,
            "updated": 1111111111111,
            "hidden": False,
            "invisible": False,
            "modifiedBy": "R: test@example.com",
            "productVersionId": 111111,
            "type": "FILE",
            "differentProductThankYouPage": None,
            "operationSystemPreference": None,
            "downloadURL": "/test/fake/openshift/downloadURL/1",
            "shortURL": "/test/fake/openshift/shortURL",
            "md5": "null",
            "sha256": None,
            "size": None,
        },
        {
            "id": 222222,
            "description": "Red Hat OpenShift Sandbox",
            "label": "Checksum File Sandbox",
            "order": 0,
            "created": 1111111111111,
            "updated": 1111111111111,
            "hidden": False,
            "invisible": False,
            "modifiedBy": "R: test@example.com",
            "productVersionId": 111111,
            "type": "FILE",
            "differentProductThankYouPage": None,
            "operationSystemPreference": None,
            "downloadURL": "/test/fake/openshift/downloadURL/2",
            "shortURL": "/test/fake/openshift/shortURL",
            "md5": "null",
            "sha256": "null",
            "size": "null",
        },
    ]
    return json


@pytest.fixture
def file_response_json():
    json = {
        "id": 111111,
        "description": "Red Hat OpenShift Local",
        "label": "Checksum File",
        "order": 0,
        "created": 1111111111111,
        "updated": 1111111111111,
        "hidden": False,
        "invisible": False,
        "modifiedBy": "R: test@example.com",
        "productVersionId": 111111,
        "type": "FILE",
        "differentProductThankYouPage": None,
        "operationSystemPreference": None,
        "downloadURL": "/test/fake/openshift/downloadURL/1",
        "shortURL": "/test/fake/openshift/shortURL",
        "md5": "null",
        "sha256": None,
        "size": None,
    }
    return json


@pytest.fixture()
def cgw_client():
    auth = CGWBasicAuth("foo", "bar")
    return CGWClient("mock://test.com/", auth)


def test_get_products_success(get_all_products_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products",
            status_code=200,
            json=get_all_products_response_json,
        )
        response = cgw_client.get_products()
        assert m.call_count == 1
        assert get_all_products_response_json == response


def test_get_products_failed(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products",
            status_code=404,
            json=False,
        )
        with pytest.raises(CGWClientError) as exception:
            cgw_client.get_products()
        assert (
            "content gateway API returned error: "
            "\nstatus_code: 404, "
            "reason: None, "
            "error: false" in str(exception.value.message)
        )
        assert m.call_count == 1


def test_get_single_product_success(product_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1",
            status_code=200,
            json=product_response_json,
        )
        response = cgw_client.get_product(1)
        assert m.call_count == 1
        assert product_response_json == response


def test_create_product_success(create_product_data, product_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "PUT",
            "mock://test.com/products/",
            status_code=200,
            json=product_response_json,
        )
        response = cgw_client.create_product(create_product_data["metadata"])
        assert m.call_count == 1
        assert product_response_json == response


def test_update_product_success(update_product_data, product_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "POST",
            "mock://test.com/products",
            status_code=200,
            json=product_response_json,
        )
        response = cgw_client.update_product(update_product_data["metadata"])
        assert m.call_count == 1
        assert product_response_json == response


def test_delete_product_success(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "DELETE",
            "mock://test.com/products/1",
            status_code=200,
        )
        cgw_client.delete_product(1)
        assert m.call_count == 1


def test_get_versions_success(get_all_versions_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions",
            status_code=200,
            json=get_all_versions_response_json,
        )
        response = cgw_client.get_versions(1)
        assert m.call_count == 1
        assert get_all_versions_response_json == response


def test_get_version_failed(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions",
            status_code=404,
            json=False,
        )
        with pytest.raises(CGWClientError) as exception:
            cgw_client.get_versions(1)
        assert (
            "content gateway API returned error: "
            "\nstatus_code: 404, "
            "reason: None, "
            "error: false" in str(exception.value.message)
        )
        assert m.call_count == 1


def test_get_single_version_success(version_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1",
            status_code=200,
            json=version_response_json,
        )
        response = cgw_client.get_version(1, 1)
        assert m.call_count == 1
        assert version_response_json == response


def test_create_version_success(version_response_json, create_version_data, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "PUT",
            "mock://test.com/products/1/versions/",
            status_code=200,
            json=version_response_json,
        )
        response = cgw_client.create_version(1, create_version_data["metadata"])
        assert m.call_count == 1
        assert version_response_json == response


def test_update_version_success(version_response_json, create_version_data, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "POST",
            "mock://test.com/products/1/versions",
            status_code=200,
            json=version_response_json,
        )
        response = cgw_client.update_version(1, create_version_data["metadata"])
        assert m.call_count == 1
        assert version_response_json == response


def test_delete_version_success(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "DELETE",
            "mock://test.com/products/1/versions/1",
            status_code=200,
        )
        cgw_client.delete_version(1, 1)
        assert m.call_count == 1


def test_get_files_success(get_all_files_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/files",
            status_code=200,
            json=get_all_files_response_json,
        )
        response = cgw_client.get_files(1, 1)
        assert m.call_count == 1
        assert get_all_files_response_json == response


def test_get_all_files_success(get_all_files_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/all",
            status_code=200,
            json=get_all_files_response_json,
        )
        response = cgw_client.get_all_files(1, 1)
        assert m.call_count == 1
        assert get_all_files_response_json == response


def test_get_files_failed(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/files/1",
            status_code=404,
            json=False,
        )
        with pytest.raises(CGWClientError) as exception:
            cgw_client.get_file(1, 1, 1)
        assert (
            "content gateway API returned error: "
            "\nstatus_code: 404, "
            "reason: None, "
            "error: false" in str(exception.value.message)
        )
        assert m.call_count == 1


def test_get_single_file_success(file_response_json, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/files/1",
            status_code=200,
            json=file_response_json,
        )
        response = cgw_client.get_file(1, 1, 1)
        assert m.call_count == 1
        assert file_response_json == response


def test_create_file_success(file_response_json, create_file_data, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "PUT",
            "mock://test.com/products/1/versions/1/files",
            status_code=200,
            json=file_response_json,
        )
        response = cgw_client.create_file(1, 1, create_file_data["metadata"])
        assert m.call_count == 1
        assert file_response_json == response


def test_update_file_success(file_response_json, create_version_data, cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "POST",
            "mock://test.com/products/1/versions/1/files",
            status_code=200,
            json=file_response_json,
        )
        response = cgw_client.update_file(1, 1, create_version_data["metadata"])
        assert m.call_count == 1
        assert file_response_json == response


def test_delete_file_success(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "DELETE",
            "mock://test.com/products/1/versions/1/files/1",
            status_code=200,
        )
        cgw_client.delete_file(1, 1, 1)
        assert m.call_count == 1


def test_get_urls_success(cgw_client):
    test_url = {"url": "test/url"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/urls",
            status_code=200,
            json=test_url,
        )
        response = cgw_client.get_urls(1, 1)
        assert m.call_count == 1
        assert response == test_url


def test_get_urls_failed(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/urls",
            status_code=404,
            json=False,
        )
        with pytest.raises(CGWClientError) as exception:
            cgw_client.get_urls(1, 1)
        assert (
            "content gateway API returned error: "
            "\nstatus_code: 404, "
            "reason: None, "
            "error: false" in str(exception.value.message)
        )
        assert m.call_count == 1


def test_get_single_url_success(cgw_client):
    test_url = {"url": "test/url"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/urls/1",
            status_code=200,
            json=test_url,
        )
        response = cgw_client.get_url(1, 1, 1)
        assert m.call_count == 1
        assert response == test_url


def test_create_url_success(cgw_client):
    test_url = {"url": "test/url"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "PUT",
            "mock://test.com/products/1/versions/1/urls/",
            status_code=200,
            json=test_url,
        )
        response = cgw_client.create_url(1, 1, test_url)
        assert m.call_count == 1
        assert response == test_url


def test_update_url_success(cgw_client):
    test_url = {"url": "test/url"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "POST",
            "mock://test.com/products/1/versions/1/urls",
            status_code=200,
            json=test_url,
        )
        response = cgw_client.update_url(1, 1, test_url)
        assert m.call_count == 1
        assert response == test_url


def test_delete_url_success(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "DELETE",
            "mock://test.com/products/1/versions/1/urls/1",
            status_code=200,
        )
        cgw_client.delete_url(1, 1, 1)
        assert m.call_count == 1


def test_get_internals_success(cgw_client):
    test_internals = {"internals": "test/internals"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/internals",
            status_code=200,
            json=test_internals,
        )
        response = cgw_client.get_internals(1, 1)
        assert m.call_count == 1
        assert response == test_internals


def test_get_internals_failed(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/internals",
            status_code=404,
            json=False,
        )
        with pytest.raises(CGWClientError) as exception:
            cgw_client.get_internals(1, 1)
        assert (
            "content gateway API returned error: "
            "\nstatus_code: 404, "
            "reason: None, "
            "error: false" in str(exception.value.message)
        )
        assert m.call_count == 1


def test_get_single_internals_success(cgw_client):
    test_internals = {"internals": "test/internals"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "mock://test.com/products/1/versions/1/internals/1",
            status_code=200,
            json=test_internals,
        )
        response = cgw_client.get_internal(1, 1, 1)
        assert m.call_count == 1
        assert response == test_internals


def test_create_internals_success(cgw_client):
    test_internals = {"internals": "test/internals"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "PUT",
            "mock://test.com/products/1/versions/1/internals/",
            status_code=200,
            json=test_internals,
        )
        response = cgw_client.create_internal(1, 1, test_internals)
        assert m.call_count == 1
        assert response == test_internals


def test_update_internals_success(cgw_client):
    test_internals = {"internals": "test/internals"}
    with requests_mock.Mocker() as m:
        m.register_uri(
            "POST",
            "mock://test.com/products/1/versions/1/internals/",
            status_code=200,
            json=test_internals,
        )
        response = cgw_client.update_internal(1, 1, test_internals)
        assert m.call_count == 1
        assert response == test_internals


def test_delete_internals_success(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "DELETE",
            "mock://test.com/products/1/versions/1/internals/1",
            status_code=200,
        )
        cgw_client.delete_internal(1, 1, 1)
        assert m.call_count == 1


def test_invalid_http_method_call(cgw_client):
    with requests_mock.Mocker() as m:
        m.register_uri(
            "INVALID_METHOD",
            "mock://test.com/products",
            json={"error": "An error has occurred!"},
        )
        with pytest.raises(CGWClientError) as exception:
            cgw_client.call_cgw_api("INVALID_METHOD", "/products")
        assert "Wrong request method passed" in str(exception.value.message)
        assert m.call_count == 0


def test_cgw_error_response():
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "/products",
            status_code=400,
            json={"error": "An error has occurred!"},
        )
        cgw_client = CGWClient("fake-host")
        with pytest.raises(CGWClientError):
            cgw_client.get_products()


def test_cgw_error_with_empty_host():
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "/products",
            status_code=400,
            json={"error": "An error has occurred!"},
        )
        with pytest.raises(CGWClientError) as exception:
            CGWClient(None)
        assert "No content gateway hostname found" in str(exception.value.message)



from pubtools._content_gateway.push_staged_cgw import PushStagedCGW, entry_point
import pytest
from tests.fake_cgw_client import TestClient
import os
from pushsource import CGWPushItem, FilePushItem, DirectoryPushItem
import json

from tests.conftest import test_staging_dir, test_staging_dir_raw

try:
    import mock
except ImportError:
    from unittest import mock

test_dir = os.path.dirname(__file__)
yaml_file_path = "test_data/test_cgw_push_staged.yaml"
invalid_push_item_path_yaml_file = "test_data/test_invalid_cgw_push_staged.yaml"


@pytest.fixture()
def target_setting():
    return {
        "target_user": "foo",
        "target_password": "bar",
        "target_address": "http://fake/server/",
    }


class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def get_pulp_push_item():
    pulp_push_item = {
        "repository_memberships": "test",
        "content_type_id": "test",
        "cdn_published": "test",
        "cdn_path": "test",
        "display_order": "test",
        "version": "v1",
        "description": "test",
        "sha256sum": "test",
        "md5sum": "test",
        "size": "test",
        "path": "test",
    }

    return Struct(**pulp_push_item)


@mock.patch("pubtools._content_gateway.push_base.CGWClient", return_value=TestClient())
@pytest.mark.parametrize(
    "fixture_source_stage",
    [
        [
            FilePushItem(
                name="dummy-1.0.0-0.x86_64.rpm",
                src="%s/repo1/RPMS/dummy-1.0.0-0.x86_64.rpm" % test_staging_dir(),
                origin="",
            ),
            CGWPushItem(
                name="cgw_push.yaml",
                src="%s/repo1/CGW/cgw.yaml" % test_staging_dir(),
                origin="",
            ),
        ]
    ],
    indirect=True,
)
def test_cgw_operations_success(mocked_cgw_client, target_setting, fixture_source_stage):
    pulp_push_item = get_pulp_push_item()
    push_cgw = PushStagedCGW(["staged:%s" % test_staging_dir()], "fake_target_name", target_setting)
    for item in push_cgw.push_items:
        push_cgw.pulp_push_units[push_cgw.push_item_str(item)] = pulp_push_item
        push_cgw.processed_push_items[push_cgw.push_item_str(item)] = item
    push_cgw.push_staged_operations()

    assert len(push_cgw.cgw_client.create_product.calls) >= 1
    assert len(push_cgw.cgw_client.create_version.calls) >= 1
    assert len(push_cgw.cgw_client.create_file.calls) >= 1
    assert mocked_cgw_client.called is True


@mock.patch("pubtools._content_gateway.push_base.CGWClient", return_value=TestClient())
@pytest.mark.parametrize(
    "fixture_source_stage",
    [
        [
            DirectoryPushItem(
                name="origin",
                src="%s/origin/RAW" % test_staging_dir_raw(),
                origin="",
            ),
            CGWPushItem(
                name="cgw_push.yaml",
                src="%s/origin/CGW/cgw.yaml" % test_staging_dir_raw(),
                origin="",
            ),
        ],
    ],
    indirect=True,
)
def test_cgw_operations_success_raw(mocked_cgw_client, target_setting, fixture_source_stage):
    push_cgw = PushStagedCGW(["staged:%s" % test_staging_dir_raw()], "fake_target_name", target_setting)
    push_cgw.push_staged_operations()

    assert len(push_cgw.cgw_client.create_product.calls) >= 1
    assert len(push_cgw.cgw_client.create_version.calls) >= 1
    assert len(push_cgw.cgw_client.create_file.calls) >= 1
    assert mocked_cgw_client.called is True


@mock.patch("pubtools._content_gateway.push_base.CGWClient", return_value=TestClient())
@pytest.mark.parametrize(
    "fixture_source_stage",
    [
        [
            FilePushItem(
                name="dummy-1.0.0-0.x86_64.rpm",
                src="%s/repo_invalid/RPMS/dummy-1.0.0-0.x86_64.rpm" % test_staging_dir(),
                origin="",
            ),
            CGWPushItem(
                name="cgw_push.yaml",
                src="%s/repo1/CGW/cgw.yaml" % test_staging_dir(),
                origin="",
            ),
        ]
    ],
    indirect=True,
)
def test_invalid_push_items(mocked_cgw_client, target_setting, fixture_source_stage):
    pulp_push_item = get_pulp_push_item()
    push_cgw = PushStagedCGW(["stage:%s" % test_staging_dir()], "fake_target_name", target_setting)
    for item in push_cgw.push_items:
        push_cgw.processed_push_items[json.dumps(repr(item), sort_keys=True)] = item
        push_cgw.pulp_push_units[json.dumps(repr(item), sort_keys=True)] = pulp_push_item

    push_cgw.rollback_cgw_operation = mock.MagicMock()

    with pytest.raises(ValueError) as exception:
        push_cgw.push_staged_operations()

    assert "Unable to find push item with" in str(exception.value)


@mock.patch("pubtools._content_gateway.push_staged_cgw.PushStagedCGW")
def test_entry_point_success(mock_push_cgw, target_setting):
    res = entry_point([], target_setting, "fake_target_name")
    assert res == mock_push_cgw.return_value
    assert mock_push_cgw.call_count == 1
    assert mock_push_cgw.called is True


@pytest.mark.parametrize(
    "fixture_source_stage",
    [
        [
            FilePushItem(
                name="dummy-1.0.0-0.x86_64.rpm",
                src="%s/repo_invalid/RPMS/dummy-1.0.0-0.x86_64.rpm" % test_staging_dir(),
                origin="",
            ),
            CGWPushItem(
                name="cgw_push.yaml",
                src="%s/repo1/CGW/cgw.yaml" % test_staging_dir(),
                origin="",
            ),
        ]
    ],
    indirect=True,
)
def test_pulp_item_push_finished_success(target_setting, fixture_source_stage):
    cgw_item = FilePushItem(
        name="dummy-1.0.0-0.x86_64.rpm",
        src="%s/repo_invalid/RPMS/dummy-1.0.0-0.x86_64.rpm" % test_staging_dir(),
        origin="",
    )

    pulp_push_item = get_pulp_push_item()
    push_cgw = PushStagedCGW(["stage:%s" % test_staging_dir()], "fake_target_name", target_setting)

    push_cgw.pulp_item_push_finished([pulp_push_item], cgw_item)
    assert push_cgw.push_items == [
        FilePushItem(
            name="dummy-1.0.0-0.x86_64.rpm",
            src="%s/repo_invalid/RPMS/dummy-1.0.0-0.x86_64.rpm" % test_staging_dir(),
            origin="",
        ),
        CGWPushItem(
            name="cgw_push.yaml",
            src="%s/repo1/CGW/cgw.yaml" % test_staging_dir(),
            origin="",
        ),
    ]
    assert push_cgw.push_item_str(cgw_item) in push_cgw.pulp_push_units



import json
from .cgw_session import CGWSession


class CGWClientError(Exception):
    """Custom exception class to handle Content Gateway Client errors"""

    def __init__(self, message):
        self.message = message


class CGWClient:
    """Class for performing content gateway HTTP API operations."""

    def __init__(self, hostname, cgw_auth=None, verify=True):
        """
        Initialize.

        Args:
            hostname (str):
                CGW host URL.
            cgw_auth (CGWBasicAuth):
                CGWBasicAuth subclass instance
            verify (bool)
                enable/disable SSL verification

        """

        self.hostname = hostname
        # Check if CGW hostname is present
        if not self.hostname:
            raise CGWClientError("No content gateway hostname found")
        self.hostname = hostname.rstrip("/")
        self.cgw_session = CGWSession(self.hostname, verify=verify)
        if cgw_auth:
            cgw_auth.make_auth(self.cgw_session)

    def call_cgw_api(self, method, endpoint, data=None):
        """
        Perform an HTTP API request on content gateway registry.

        Args:
            method (str):
                REST API method of the request (GET, POST, PUT, DELETE).
            endpoint (str):
                Endpoint of the request.
            data (dict):
                Optional arguments for the method Request object.

        Returns (list|dict|int):
            Returns either list or dict response for GET method
            and int for PUT methods.

        Raises:
            CGWClientError: When the request returns an error status.
            Exception: Occurs if API call fails.
        """

        # Check if correct method id passed
        if method not in ["GET", "POST", "PUT", "DELETE"]:
            raise CGWClientError("Wrong request method passed")

        response = None
        output = {}
        try:
            if method == "GET":
                response = self.cgw_session.get(endpoint, data=data)
            elif method == "POST":
                response = self.cgw_session.post(endpoint, data=data)
            elif method == "PUT":
                response = self.cgw_session.put(endpoint, data=data)
            elif method == "DELETE":
                response = self.cgw_session.delete(endpoint)
        except Exception as e:
            msg = "Error: Exception occurred during API call: %s" % str(e)
            raise CGWClientError(msg)

        if not response.ok:
            raise CGWClientError(
                "content gateway API returned error: "
                "\nstatus_code: %s, reason: %s, error: %s" % (response.status_code, response.reason, response.text)
            )

        if response.text:
            output = response.json()
        return output

    def get_products(self):
        """
        Get list of all products.

        Returns (list[dict]):
            Returns list of products.
        """

        resp_data = self.call_cgw_api("GET", "/products")
        return resp_data

    def get_product(self, pid):
        """
        Get one product record specified by its identifier.

        Args:
            pid (int):
                product id of requested product

        Returns (dict):
            Returns the requested product.
        """

        endpoint = "/products/%s" % pid
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def create_product(self, data):
        """
        Creates a new product record on the content gateway.

        Args:
            data (dict):
                product data to be created

        Returns (int):
            Returns the created product id.
        """

        endpoint = "/products/"
        resp_data = self.call_cgw_api("PUT", endpoint, data=json.dumps(data))
        return resp_data

    def update_product(self, data):
        """
        Updates an existing product record.

        Args:
            data (dict):
                product data to be updated

        Returns (dict):
            Returns empty dict on successful product update.
        """

        endpoint = "/products"
        resp_data = self.call_cgw_api("POST", endpoint, data=json.dumps(data))
        return resp_data

    def delete_product(self, pid):
        """
        Attempts to delete an existing product record.
        It is not possible to delete products,
        that has any product versions.

        Args:
            pid (int):
                product id of the record that need to be deleted

        Returns (dict):
            Returns empty dict on successful product delete.
        """

        endpoint = "/products/%s" % pid
        resp_data = self.call_cgw_api("DELETE", endpoint)
        return resp_data

    def get_versions(self, pid):
        """
        Get list of versions belonging to the specific product.

        Args:
            pid (int):
                product id of the version

        Returns (list[dict]):
            Returns a list of a specific product versions
        """

        endpoint = "/products/%s/versions" % pid
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def get_version(self, pid, vid):
        """
        Get one product version record specified by its identifier.

        Args:
            pid (int):
                product id of the version
            vid (int):
                requested version id

        Returns (dict):
           A single version record.
        """

        endpoint = "/products/%s/versions/%s" % (pid, vid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def create_version(self, pid, data):
        """
        Creates a new product version record on the content gateway.

        Args:
            pid (int):
                product id for which version needs to be created
            data (dict):
                version data to be created

        Returns (int):
            Returns the created version id.
        """
        endpoint = "/products/%s/versions/" % pid
        resp_data = self.call_cgw_api("PUT", endpoint, data=json.dumps(data))
        return resp_data

    def update_version(self, pid, data):
        """
        Updates an existing version record.

        Args:
            pid (int):
                product id of the version
            data (dict):
                version data to be updated

        Returns (dict):
            Returns empty dict on successful version update.
        """

        endpoint = "/products/%s/versions" % pid
        resp_data = self.call_cgw_api("POST", endpoint, data=json.dumps(data))
        return resp_data

    def delete_version(self, pid, vid):
        """
        Attempts to delete an existing version record.
        It is not possible to delete a version,
        that has any file records.

        Args:
            pid (int):
                product id of the version
            vid (int):
                version id of the record that need to be deleted

        Returns (dict):
            Returns empty dict on successful version delete.
        """

        endpoint = "/products/%s/versions/%s" % (pid, vid)
        resp_data = self.call_cgw_api("DELETE", endpoint)
        return resp_data

    def get_all_files(self, pid, vid):
        """
        Get list of all files, links and internal links of the specific version.

        Args:
            pid (int):
                product id of the file
            vid (int):
                version id of the file

        Returns (list[dict]):
            Returns a list of a specific product version files
        """

        endpoint = "/products/%s/versions/%s/all" % (pid, vid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def get_files(self, pid, vid):
        """
        Get list of files belonging to the specific product version.

        Args:
            pid (int):
                product id of the file
            vid (int):
                version id of the file

        Returns (list[dict]):
            Returns a list of a specific product version files
        """

        endpoint = "/products/%s/versions/%s/files" % (pid, vid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def get_file(self, pid, vid, fid):
        """
        Get a single file specified by its identifier.

        Args:
            pid (int):
                product id of the file
            vid (int):
                version id of the file
            fid (int):
                requested file id

        Returns (dict):
           A single file record.
        """

        endpoint = "/products/%s/versions/%s/files/%s" % (pid, vid, fid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def create_file(self, pid, vid, data):
        """
        Creates a new file record on the content gateway.

        Args:
            pid (int):
                product id of the file
            vid (int):
                version id for which file needs to be created
            data (dict):
                file data to be created

        Returns (int):
            Returns the created file id.
        """

        endpoint = "/products/%s/versions/%s/files" % (pid, vid)
        resp_data = self.call_cgw_api("PUT", endpoint, data=json.dumps(data))
        return resp_data

    def update_file(self, pid, vid, data):
        """
        Updates an existing file record.

        Args:
            pid (int):
                product id of the file
            vid (int):
                version id for which file needs to be updated
            data (dict):
                file data to be updated

        Returns (dict):
            Returns empty dict on successful file update.
        """

        endpoint = "/products/%s/versions/%s/files" % (pid, vid)
        resp_data = self.call_cgw_api("POST", endpoint, data=json.dumps(data))
        return resp_data

    def delete_file(self, pid, vid, fid):
        """
        Attempts to delete an existing file record.

        Args:
            pid (int):
                product id of the file
            vid (int):
                version id of the file
            fid (int):
                file id of the record that need to be deleted

        Returns (dict):
            Returns empty dict on successful file delete.
        """

        endpoint = "/products/%s/versions/%s/files/%s" % (pid, vid, fid)
        resp_data = self.call_cgw_api("DELETE", endpoint)
        return resp_data

    def get_urls(self, pid, vid):
        """
        Get list of urls belonging to the specific product version.

        Args:
            pid (int):
                product id of the url
            vid (int):
                version id of the url

        Returns (list[dict]):
            Returns a list of a specific product version urls
        """

        endpoint = "/products/%s/versions/%s/urls" % (pid, vid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def get_url(self, pid, vid, uid):
        """
        Get a single url specified by its identifier.

        Args:
            pid (int):
                product id of the url
            vid (int):
                version id of the url
            uid (int):
                requested url id

        Returns (dict):
           A single url record.
        """

        endpoint = "/products/%s/versions/%s/urls/%s" % (pid, vid, uid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def create_url(self, pid, vid, data):
        """
        Creates a new url record on the content gateway.

        Args:
            pid (int):
                product id of the url
            vid (int):
                version id for which url needs to be created
            data (dict):
                url data to be created

        Returns (int):
            Returns the created url id.
        """

        endpoint = "/products/%s/versions/%s/urls/" % (pid, vid)
        resp_data = self.call_cgw_api("PUT", endpoint, data=json.dumps(data))
        return resp_data

    def update_url(self, pid, vid, data):
        """
        Updates an existing url record.

        Args:
            pid (int):
                product id of the url
            vid (int):
                version id for which url needs to be updated
            data (dict):
                url data to be updated

        Returns (dict):
            Returns empty dict on successful url update.
        """

        endpoint = "/products/%s/versions/%s/urls" % (pid, vid)
        resp_data = self.call_cgw_api("POST", endpoint, data=json.dumps(data))
        return resp_data

    def delete_url(self, pid, vid, uid):
        """
        Attempts to delete an existing url record.

        Args:
            pid (int):
                product id of the version
            vid (int):
                version id of the url
            uid (int):
                url id of the record that need to be deleted

        Returns (dict):
            Returns empty dict on successful url delete.
        """

        endpoint = "/products/%s/versions/%s/urls/%s" % (pid, vid, uid)
        resp_data = self.call_cgw_api("DELETE", endpoint)
        return resp_data

    def get_internals(self, pid, vid):
        """
        Get list of internals belonging to the specific product version.

        Args:
            pid (int):
                product id of the internals
            vid (int):
                version id of the internals

        Returns (list[dict]):
            Returns a list of a specific product version internals
        """

        endpoint = "/products/%s/versions/%s/internals" % (pid, vid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def get_internal(self, pid, vid, iid):
        """
        Get a single internal specified by its identifier.

        Args:
            pid (int):
                product id of the internal
            vid (int):
                version id of the internal
            iid (int):
                requested internal id

        Returns (dict):
           A single internal record.
        """

        endpoint = "/products/%s/versions/%s/internals/%s" % (pid, vid, iid)
        resp_data = self.call_cgw_api("GET", endpoint)
        return resp_data

    def create_internal(self, pid, vid, data):
        """
        Creates a new internal record on the content gateway.

        Args:
            pid (int):
                product id of the internal
            vid (int):
                version id for which internal needs to be created
            data (dict):
                internal data to be created

        Returns (int):
            Returns the created internal id.
        """

        endpoint = "/products/%s/versions/%s/internals/" % (pid, vid)
        resp_data = self.call_cgw_api("PUT", endpoint, data=json.dumps(data))
        return resp_data

    def update_internal(self, pid, vid, data):
        """
        Updates an existing internal record.

        Args:
            pid (int):
                product id of the internal
            vid (int):
                version id for which internal needs to be updated
            data (dict):
                internal data to be updated

        Returns (dict):
            Returns empty dict on successful internal update.
        """

        endpoint = "/products/%s/versions/%s/internals/" % (pid, vid)
        resp_data = self.call_cgw_api("POST", endpoint, data=json.dumps(data))
        return resp_data

    def delete_internal(self, pid, vid, iid):
        """
        Attempts to delete an existing internal record.

        Args:
            pid (int):
                product id of the internal
            vid (int):
                version id of the internal
            iid (int):
                url id of the record that need to be deleted

        Returns (dict):
            Returns empty dict on successful internal delete.
        """

        endpoint = "/products/%s/versions/%s/internals/%s" % (pid, vid, iid)
        resp_data = self.call_cgw_api("DELETE", endpoint)
        return resp_data



try:
    from unittest.mock import patch
except ImportError:
    from mock import patch
from pubtools._content_gateway.cgw_session import CGWSession


@patch("requests.Session.get")
@patch("requests.Session.post")
@patch("requests.Session.put")
@patch("requests.Session.delete")
def test_cgw_session_methods(patched_delete, patched_put, patched_post, patched_get):
    cgw_session = CGWSession("https://fake-host")
    cgw_session.get("/fake-end-point")
    cgw_session.post("/fake-end-point")
    cgw_session.put("/fake-end-point")
    cgw_session.delete("/fake-end-point")

    patched_get.assert_called_with("https://fake-host/fake-end-point", verify=True)
    patched_post.assert_called_with("https://fake-host/fake-end-point", verify=True)
    patched_put.assert_called_with("https://fake-host/fake-end-point", verify=True)
    patched_delete.assert_called_with("https://fake-host/fake-end-point", verify=True)



from base64 import b64encode


class CGWAuth(object):
    """Custom exception to handle authentication errors."""

    def __init__(self):
        raise NotImplementedError

    def make_auth(self, cgw_session):
        raise NotImplementedError  # pragma: no cover


class CGWClientError(Exception):
    """Custom exception to handle Content Gateway Client errors."""

    def __init__(self, message):
        self.message = message


class CGWBasicAuth(CGWAuth):
    """Basic Auth provider to CGWClient."""

    # pylint: disable=super-init-not-called
    def __init__(self, user, password):
        """Initializing

        Args:
            user (str)
                Basic auth user_name
            password (str)
                Basic auth password
        """
        self.user = user
        self.password = password

    def make_auth(self, cgw_session=None):
        """Setup CGWSession with basic auth.

        Args:
            cgw_session (CGWSession)
                CGWSession instance
        """

        if not self.user or not self.password:
            raise CGWClientError("Error: username / password not found !!")

        user_pass_string = "%s:%s" % (self.user, self.password)
        user_pass = b64encode(user_pass_string.encode("utf-8")).decode("ascii")
        cgw_session.session.headers["Authorization"] = "Basic %s" % user_pass



import os
import json
import logging
import hashlib
from .push_base import PushBase
from .utils import yaml_parser, validate_data, sort_items, format_cgw_items

try:
    import attr as attrs
except ImportError:  # pragma: no cover
    import attrs

from pubtools.pluggy import hookimpl, pm
from pushsource import CGWPushItem, DirectoryPushItem
from pushsource import Source

LOG = logging.getLogger("pubtools.cgw")
LOG_FORMAT = "%(asctime)s [%(levelname)-8s] %(message)s"


class PushStagedCGW(PushBase):
    """Handle push staged CGW workflow."""

    def __init__(self, source_urls, target_name, target_settings):
        """
        Initialize.

        Args:
            target_name (str):
                Name of the target.
            target_settings (dict):
                Target settings.
        """
        PushBase.__init__(
            self,
            target_settings["target_address"],
            target_settings["target_user"],
            target_settings["target_password"],
        )
        self.push_items = []
        self.pulp_push_units = {}
        self.processed_push_items = {}
        self.source_urls = source_urls

        for source_url in source_urls:
            with Source.get(source_url) as source:
                LOG.info("Loading items from %s", source_url)
                for item in source:
                    self.push_items.append(item)

    @staticmethod
    def push_item_str(item):
        push_item_dict = attrs.asdict(item)
        item_json = {
            "name": push_item_dict["name"],
            "src": push_item_dict["src"],
            "dest": push_item_dict["dest"],
            "origin": push_item_dict["origin"],
            "build": push_item_dict["build"],
            "signing_key": push_item_dict["signing_key"],
        }
        return json.dumps(item_json, sort_keys=True)

    @hookimpl
    def pulp_item_push_finished(self, pulp_units, push_item):
        """
        Invoked during task execution after successful completion of all Pulp
        publishes.

        This hook is invoked for every push item, to indicate that all Pulp
        content associated with the task is considered fully up-to-date. The
        intended usage is to flush Pulp-derived caches or to notify systems that
        Pulp content may have recently changed.

        Args:
            pulp_push_item (Pulp object):
                push item from the pulp.
            push_item (list):
                push item.
        """

        if pulp_units:
            self.pulp_push_units[self.push_item_str(push_item)] = pulp_units[0]
            self.processed_push_items[self.push_item_str(push_item)] = push_item

    def push_staged_operations(self):
        """
        Initiate the CGW operations for push staged.
        Operations such as create, update or delete
        on products, versions and files would be performed.
        """

        for item in self.push_items:
            if isinstance(item, CGWPushItem):
                parsed_items = yaml_parser(os.path.join(item.origin, item.src))
                parsed_items = format_cgw_items(parsed_items)

                for pitem in parsed_items:
                    validate_data(pitem)

                parsed_items = sort_items(parsed_items)
                try:
                    for pitem in parsed_items:
                        if pitem["type"] == "product":
                            self.process_product(pitem)
                        if pitem.get("type") == "product_version":
                            self.process_version(pitem)
                        if pitem["type"] == "file":
                            if "pushItemPath" in pitem["metadata"]:
                                # push to CDN and set required pitem attributes for CGW entry
                                found = None
                                for push_item in self.push_items:
                                    if isinstance(push_item, DirectoryPushItem):
                                        found = "directory_item"
                                    elif (
                                        push_item.src.replace(push_item.origin, "").lstrip("/")
                                        == pitem["metadata"]["pushItemPath"]
                                    ):
                                        found = "file_item"
                                    if found:
                                        break
                                else:
                                    raise ValueError(
                                        "Unable to find push item with path:%s" % pitem["metadata"]["pushItemPath"]
                                    )
                                if found == "file_item":
                                    pulp_push_unit = self.pulp_push_units[self.push_item_str(push_item)]
                                    pulp_push_item = self.processed_push_items[self.push_item_str(push_item)]
                                    pitem["metadata"]["downloadURL"] = pulp_push_unit.cdn_path
                                    pitem["metadata"]["md5"] = pulp_push_item.md5sum
                                    pitem["metadata"]["sha256"] = pulp_push_unit.sha256sum
                                    pitem["metadata"]["size"] = os.stat(push_item.src).st_size
                                else:
                                    filename = pitem["metadata"]["pushItemPath"].split("/")[-1]
                                    file_path = os.path.join(push_item.src, filename)
                                    with open(file_path, "rb") as f:
                                        bytes = f.read()
                                        pitem["metadata"]["md5"] = hashlib.md5(bytes).hexdigest()
                                        pitem["metadata"]["sha256"] = hashlib.sha256(bytes).hexdigest()
                                    pitem["metadata"]["size"] = os.path.getsize(file_path)
                                    pitem["metadata"]["downloadURL"] = (
                                        "/content/origin/files/sha256/{0}/{1}/{2}".format(
                                            pitem["metadata"]["sha256"][:2],
                                            pitem["metadata"]["sha256"],
                                            filename,
                                        )
                                    )
                            # carry out CGW operations
                            self.process_file(pitem)

                    self.make_visible()
                    LOG.info("\n All CGW operations are successfully completed...!")
                except Exception as error:
                    LOG.exception("Exception occurred during the CGW operation %s" % error)
                    LOG.info("Rolling back the partial operation")
                    self.rollback_cgw_operation()

                    #  raising the occurred Exception as all the exception will get caught in this except block
                    #  we want to return full Traceback
                    raise error


def entry_point(source_urls, target_name, target_settings):
    """Entrypoint for CGW push stage."""
    logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
    ret = PushStagedCGW(source_urls, target_name, target_settings)
    pm.register(ret)
    return ret



# -*- coding: utf-8 -*-
"""setup.py"""

import os
import sys
from setuptools import setup, find_namespace_packages
from setuptools.command.test import test as TestCommand


class Tox(TestCommand):
    user_options = [("tox-args=", "a", "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox
        import shlex

        if self.tox_args:
            errno = tox.cmdline(args=shlex.split(self.tox_args))
        else:
            errno = tox.cmdline(self.test_args)
        sys.exit(errno)


def read_content(filepath):
    with open(filepath) as fobj:
        return fobj.read()


long_description = read_content("docs/README.rst") + read_content(
    os.path.join("docs/", "CHANGELOG.rst")
)

extras_require = {"reST": ["Sphinx"]}
if os.environ.get("READTHEDOCS", None):
    extras_require["reST"].append("recommonmark")


def get_description():
    return "Automate pushing to Content Gateway"


def get_long_description():
    with open("docs/README.rst") as f:
        text = f.read()

    # Long description is everything after README's initial heading
    idx = text.find("\n\n")
    return text[idx:]


def get_requirements():
    # Note: at time of writing, requirements.txt is empty.
    # It exists anyway for a consistent setup.
    with open("requirements.txt") as f:
        return f.read().splitlines()


# INSTALL_REQUIRES = ['urllib3', 'six', 'requests-mock', 'requests', 'mock', 'pytest', 'setuptools']
INSTALL_REQUIRES = get_requirements()

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
]

setup(
    name="pubtools-content-gateway",
    version="0.5.4",
    description=get_description(),
    long_description_content_type="text/markdown",
    author="Javed Alam",
    author_email="jalam@redhat.com",
    url="https://github.com/release-engineering/pubtools-content-gateway",
    classifiers=classifiers,
    packages=find_namespace_packages(where="src"),
    package_dir={"": "src"},
    install_requires=INSTALL_REQUIRES,
    tests_require=["tox", "mock", "requests_mock", "pushcollector", "requests", "tox", "pytest", "covdefaults",
                   "pytest-cov"],
    data_files=[],
    entry_points={
        "console_scripts": [
            "push-cgw-metadata = pubtools._content_gateway.push_cgw:main"
        ],
        "target": [
            "push-staged-cgw = pubtools._content_gateway.push_staged_cgw:entry_point",
        ]
    },
    include_package_data=True,
    python_requires=">=3.9",
    project_urls={
        "Changelog": "https://release-engineering.github.io/pubtools-content-gateway/CHANGELOG.html",
        "Documentation": "https://release-engineering.github.io/pubtools-content-gateway/",
    },
    cmdclass={"test": Tox},
)



import yaml
from yaml.loader import SafeLoader
from jsonschema import validate
import logging
import copy

LOG = logging.getLogger("pubtools.cgw")

PRODUCT_SCHEMA = {
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "action": {"type": "string", "enum": ["create", "update", "delete"]},
        "metadata": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "productCode": {"type": ["string", "null"]},
                "homepage": {"type": "string"},
                "downloadpage": {"type": "string"},
                "thankYouPage": {"type": "string"},
                "eloquaCode": {"type": ["string", "number"]},
                "featuredArtifactType": {"type": "string"},
                "thankYouTimeout": {"type": "integer"},
            },
            "required": ["name", "productCode", "eloquaCode"],
        },
    },
    "required": ["type", "action", "metadata"],
}

VERSION_SCHEMA = {
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "action": {"type": "string", "enum": ["create", "update", "delete"]},
        "metadata": {
            "type": "object",
            "properties": {
                "productName": {"type": "string"},
                "productCode": {"type": ["string", "null", "number"]},
                "versionName": {"type": ["string", "number"]},
                "ga": {"type": "boolean"},
                "termsAndConditions": {"type": "string"},
                "trackingDisabled": {"type": "boolean"},
                "hidden": {"type": "boolean"},
                "invisible": {"type": "boolean"},
                "releaseDate": {"type": "string"},
            },
            "required": [
                "productName",
                "productCode",
                "versionName",
                "termsAndConditions",
            ],
        },
    },
    "required": ["type", "action", "metadata"],
}

FILE_SCHEMA = {
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "action": {"type": "string", "enum": ["create", "update", "delete"]},
        "metadata": {
            "oneOf": [
                {
                    "type": "object",
                    "properties": {
                        "productName": {"type": "string"},
                        "productCode": {"type": ["string", "null"]},
                        "productVersionName": {"type": ["string", "number", "null"]},
                        "description": {"type": ["string", "null"]},
                        "label": {"type": ["string", "null"]},
                        "order": {"type": "integer"},
                        "hidden": {"type": "boolean"},
                        "invisible": {"type": "boolean"},
                        "type": {"type": "string"},
                        "differentProductThankYouPage": {"type": ["number", "null"]},
                        "shortURL": {"type": "string"},
                        "pushItemPath": {"type": "string"},
                    },
                    "required": [
                        "productName",
                        "productCode",
                        "productVersionName",
                        "pushItemPath",
                    ],
                },
                {
                    "type": "object",
                    "properties": {
                        "productName": {"type": "string"},
                        "productCode": {"type": ["string", "null"]},
                        "productVersionName": {"type": ["string", "number", "null"]},
                        "description": {"type": ["string", "null"]},
                        "label": {"type": ["string", "null"]},
                        "order": {"type": "integer"},
                        "hidden": {"type": "boolean"},
                        "invisible": {"type": "boolean"},
                        "type": {"type": "string"},
                        "differentProductThankYouPage": {"type": ["number", "null"]},
                        "downloadURL": {"type": "string"},
                        "shortURL": {"type": "string"},
                        "size": {"type": ["number", "null"]},
                        "md5": {"type": ["string", "number", "null"]},
                    },
                    "required": [
                        "productName",
                        "productCode",
                        "productVersionName",
                        "downloadURL",
                    ],
                },
            ]
        },
    },
    "required": ["type", "action", "metadata"],
}


def validate_data(json_data):
    """
    Validate that json_data contains all the necessary data
    with defined with json schemas

    Args:
        json_data (dict)
            JSON dictionary with content gateway data

    Raises:
        ValidationError
            The json_data type doesn't match with schema
    Returns:
        True: if validation succeed
    """

    item_type = json_data.get("type")
    if item_type == "product":
        validate(instance=json_data, schema=PRODUCT_SCHEMA)
    elif item_type == "product_version":
        validate(instance=json_data, schema=VERSION_SCHEMA)
    elif item_type == "file":
        validate(instance=json_data, schema=FILE_SCHEMA)
    LOG.info("Data validation successful for %s: %s" % (item_type, json_data.get("metadata").get("productCode")))
    return True


def yaml_parser(file_path):
    """
    Parse the yaml data into json data

    Args:
        file_path (str)
            JSON dictionary with content gateway data
    Raises:
        FileNotFoundError
            If the file_path cannot be found
    Returns:
        list(dict): parsed json data from yaml file
    """

    with open(file_path) as f:
        data = list(yaml.load_all(f, Loader=SafeLoader))
    return data[0]


def sort_items(items):
    """
    Sort the items in the following order
        1) all products with action create/update
        2) all versions with action create/update
        3) all files with action create/update
        4) all files with action delete
        5) all versions with action delete
        6) all products with action delete

    This is needed to process all the creation and update action of products,
    versions and files, then delete action of files, versions and products

    This sorting is needed for the following reasons:
        1) A file cannot be created if it's parent version and product are not present.
            All parent data need to be present before creating child object.
        2) A product cannot get deleted if it has versions and file.
            All nested data need to be cleared before deleting parent object

    Args:
        items (list(dict))
            list of JSON dictionary
    Returns:
        list(dict): list of JSON dictionary
    """
    product_create_update = []
    version_create_update = []
    file_create_update = []
    file_delete = []
    version_delete = []
    product_delete = []
    sorted_items = []

    for data in items:
        if data["type"] == "product":
            (
                product_create_update.append(data)
                if data["action"]
                in [
                    "create",
                    "update",
                ]
                else product_delete.append(data)
            )

        if data["type"] == "product_version":
            (
                version_create_update.append(data)
                if data["action"]
                in [
                    "create",
                    "update",
                ]
                else version_delete.append(data)
            )

        if data["type"] == "file":
            (
                file_create_update.append(data)
                if data["action"]
                in [
                    "create",
                    "update",
                ]
                else file_delete.append(data)
            )

    sorted_items.extend(product_create_update)
    sorted_items.extend(version_create_update)
    sorted_items.extend(file_create_update)
    sorted_items.extend(file_delete)
    sorted_items.extend(version_delete)
    sorted_items.extend(product_delete)

    return sorted_items


def format_cgw_items(items):
    """
    The yaml file can accept both linear and nested data structure
    and the function will re-format the yaml nested data structure to linear data structure
    inorder to process.

    Args:
        items (list(dict))
            list of JSON dictionary
    Returns:
        list(dict): list of JSON dictionary
    """

    formatted_list = list()
    for product_rec in items:
        if product_rec.get("type"):
            """
            Checking whether the cgw data structure is nested or linear.
                - For the linear structure "type", "action" and "metadata" keys will be present
                - If data is of linear type, we are appending the record directly in the formatted_list list
                - Else we are re-formatting nested data into linear data in the outer scope of the "if" condition.
            """
            formatted_list.append(product_rec)
            continue

        # creating a temp record of product data
        product_metadata = copy.deepcopy(product_rec["product"])
        # this action value will be shared crossed to child records i.e versions and files
        action = product_metadata["action"]
        payload = {"type": "product", "action": action}
        # performing further operations on temp record
        # i.e "product_metadata" to keep original data intact
        # and removing unnecessary fields from the temp product_metadata to form expected metadata
        product_metadata.pop("releases", None)
        product_metadata.pop("action", None)
        # expected metadata is ready, adding to the main dict
        payload["metadata"] = product_metadata
        product_name = product_metadata["name"]
        product_code = product_metadata["productCode"]
        # nested product data got converted to linear structure
        formatted_list.append(payload)  # adding record to the result list

        # re-formatting nested versions records in linear order
        # version follows the same steps as product
        # to convert sequential structure to linear record
        if product_rec.get("product").get("releases"):
            for version_rec in product_rec.get("product").get("releases"):
                # version shares the same action value as product
                version_payload = {"type": "product_version", "action": action}
                version_metadata = copy.deepcopy(version_rec)
                version_payload["metadata"] = version_metadata
                version_payload["metadata"]["productName"] = product_name
                version_payload["metadata"]["productCode"] = product_code
                version_name = version_metadata["versionName"]
                version_metadata.pop("files", None)
                formatted_list.append(version_payload)

                order = 0
                # re-formatting nested file records in linear order
                # file follows the same steps as product and version
                # to convert sequential structure to linear record
                if version_rec.get("files"):
                    for file_rec in version_rec.get("files"):
                        # file shares the same action value as product
                        file_payload = {"type": "file", "action": action}
                        order = file_rec.get("order") if file_rec.get("order") is not None else order + 10
                        file_rec["order"] = file_rec.get("order") if file_rec.get("order") else order
                        file_payload["metadata"] = file_rec
                        file_payload["metadata"]["productName"] = product_name
                        file_payload["metadata"]["productCode"] = product_code
                        file_payload["metadata"]["productVersionName"] = version_name
                        formatted_list.append(file_payload)

    return formatted_list
